

#local


#code PAGE0

; pseudo random number generator seed
; must not be zero
rndseed16:: dw	$4711		


#code VCC_INT
#include "vars16.s"



; ______________________________________
; RSWAPWITHVAR ( T T* -- T )

RSWAPWITHVAR::
	ex	hl,de		; hl -> var
	pop	de		; meine var
	
rswv:	ld	a,(hl)
	ld	(hl++),e
	ld	e,a
	
	ld	a,(hl)
	ld	(hl),d
	ld	d,a
	
	jp	next

; ______________________________________
; NSWAPWITHVAR ( T* T -- T )

NSWAPWITHVAR::
	pop	hl		; hl -> var
	jr	rswv




; -----------------------------------------
; Arrays and Structs:
; -----------------------------------------
;
; variables are pointers to handles
; handles are pointers to payload data
; struct / array length.w is at payload-2




; -----------------------------------------
; Arithmetics & Logics:
; -----------------------------------------

; ________________________________
; increment value
;
; INCR ( n -- n )

INCR::	inc	de
	jp	next

; ________________________________
; decrement value
;
; DECR ( n -- n )

DECR::	dec	de
	jp	next

; ________________________________
; calculate absolute value of signed int
; return value is uint16, because abs($8000) = $8000
;
; ABS ( n -- n )

ABS::	bit	7,d
	jp	z,next	; positive
	;jr	NEG

; ________________________________
; negate
;
; NEG ( n -- n )

NEG::	xor	a
	sub	a,e
	ld	e,a
	sbc	a,a	; nc => 0, cy => 0xff
	sub	a,d
	ld	d,a
	next

; ________________________________
; complement
;
; CPL ( n -- n )

CPL::	ld	a,d
	cpl
	ld	d,a
	ld	a,e
	cpl
	ld	e,a
	next

; ________________________________
; boolean negation: !n
;
; NOT ( n -- n )

NOT::	ld	a,d
	or	e
	ld	de,0
	jr	nz,3$
	inc	e
3$:	next

; ________________________________
; calculate sign of signed int
;
; SIGN ( n -- n )

SIGN::	ld	a,d
	add	a
	jr	c,1$	; -1
	or	e
	jr	z,2$	; 0

	ld	de,1
2$:	jp	next

1$:	ld	de,-1
	jp	next

; ________________________________
; convert to boolean
;
; BOOL ( n -- n )

BOOL::	ld	a,d
	or	e
	jr	z,3$
	ld	de,1
3$:	next

; ________________________________
; shift left once
;
; SL1 ( n -- n )

SL1::	ex	de,hl
	add	hl,hl
	ex	de,hl
	next

; ________________________________
; unsigned shift right two digits
;
; SR2u ( n -- n )

SR2u::	srl	d
	rr	e
	;jr	SR1u
	
; ________________________________
; unsigned shift right once
;
; SR1u ( n -- n )

SR1u::	srl	d
	rr	e
	next

; ________________________________
; subtract two word pointers
;
; SUBp16i ( w1* w2* -- n )	

SUBp16p::
	ex	hl,de		; de = p2
	pop	hl		; hl = p1
	and	a
	sbc	hl,de		; hl = p1-p2
	ex	de,hl		; de = p1-p2
	;jp	SR1s

; ________________________________
; signed shift right once
;
; SR1s ( n -- n )

SR1s::	sra	d
	rr	e
	next

; ________________________________
; add
;
; ADD    ( n n -- n )
; ADDp8i ( b* n -- b* )		; add offset to byte pointer

ADDp8i::
ADD::	pop	hl
	add	hl,de
	ex	de,hl
	next

; ________________________________
; subtract n1 - n2
;
; SUB    ( n1 n2 -- n )
; SUBp8p ( b* b* -- n )		; subtract two byte pointers
; SUBp8i ( b* n -- b* )		; subtract offset from byte pointer

SUBp8p::
SUBp8i::
SUB::	pop	hl		; n1
	and	a
	sbc	hl,de
	ex	de,hl
	next

; ________________________________
; add offset to word pointer
;
; ADDp16i ( w* n -- w* )

ADDp16i::
	ex	hl,de
	add	hl,hl		; hl = offset*2
	pop	de
	add	hl,de
	ex	de,hl
	jp	next

; ________________________________
; subtract offset from word pointer
;
; SUBp16i ( w* n -- w* )	

SUBp16i::
	ex	hl,de
	add	hl,hl		
	ex	hl,de		; de = offset*2
	pop	hl
	and	a
	sbc	hl,de
	ex	de,hl
	jp	next

; ________________________________
; add immediate byte 0x0000 .. 0x00FF
;
; ADDI ( n $b -- n )

ADDIu8::
	ld	a,(bc++)
	add	e
	ld	e,a
	jr	nc,3$
	inc	d
3$:	next

; ________________________________
; add immediate byte 0xFF00 .. 0xFFFF
;
; ADDIs8 ( n $b -- n )

ADDIs8::	
	ld	a,(bc++)
	add	e
	ld	e,a
	jr	c,3$
	dec	d
3$:	next

; ________________________________
; bitwise OR
;
; OR ( n n -- n )

OR::	pop	hl
	ld	a,h
	or	d
	ld	d,a
	ld	a,l
	or	e
	ld	e,a
	jp	next

; ________________________________
; bitwise XOR
;
; XOR ( n n -- n )

XOR::	pop	hl
	ld	a,h
	xor	d
	ld	d,a
	ld	a,l
	xor	e
	ld	e,a
	jp	next

; ________________________________
; bitwise AND
;
; AND ( n n -- n )

AND::	pop	hl
	ld	a,h
	and	d
	ld	d,a
	ld	a,l
	and	e
	ld	e,a
	jp	next

; ________________________________
; get high-byte of word
;
; GETHI ( n -- n )

GETHI::	ld	e,d
;	jp	GETLO

; ________________________________
; get low-byte of word
;
; GETLO ( n -- n )

GETLO::	ld	d,0
	next

; ________________________________
; swap high and low byte
;
; SWAPBYTES ( n -- n )

SWAPBYTES::
	ld	a,d
	ld	d,e
	ld	e,a
	next

; ________________________________
; calculate unsigned maximum
;
; MAXu ( n n -- n )

MAXu::	pop	hl
	and	a
	sbc	hl,de		
	jr	c,4$		; hl-de<0 => hl<de => max=de => fertig
	add	hl,de		; else: max=hl => hl restaurieren
	ex	de,hl		; 	  und nach de laden
4$:	jp	next

; ________________________________
; calculate unsigned minimum
;
; MINu ( n n -- n )

MINu::	pop	hl
	and	a
	sbc	hl,de
	jr	nc,4$		; hl-de≥0 => hl≥de => min=de => fertig
	add	hl,de		; else: min=hl => hl restaurieren
	ex	de,hl		; 	  und nach de laden
4$:	jp	next

; ________________________________
; calculate the most significant bit:
; the bit number of the highest set '1' bit
; returns 15 .. 0
; returns 0 for $0001 and for $0000
;
; MSBIT ( n -- n )

MSBIT::	ld	a,d
	and	a
	jr	nz,1$		; falls ein Bit in uint16.hi gesetzt

	ld	a,e
	and	a
	jr	z,4$		; msbit(0) = 0

	ld	de,7		; msbit = max. 7 wenn ein Bit in uint16.lo gesetzt
	jr	2$

1$:	ld	de,16		; msbit = max. 15 wenn ein Bit in uint16.hi gesetzt

3$:	dec	e
2$:	rla
	jr	nc,3$
4$:	next

; ________________________________
; shift left n1<<n2
;
; SL ( n1 n2 -- n )

SL::	ld	hl,exhlde_next
	ex	hl,(sp)		; hl = n1, de = shift
	;jp	sl
	
; ________________________________
; shift left
;
; sl:  hl = hl << e

sl::	ld	a,e
	and	a,$F
	ret	z

	cp	a,8
	jr	c,2$

	ld	h,l
	ld	l,0
	ret	z
	sub	a,8

2$:	add	hl,hl
	dec	a
	jr	nz,2$
	ret

; ________________________________
; unsigned shift right n1>>n2
;
; SRu ( n1 n2 -- n )

SRu::	ld	hl,exhlde_next
	ex	hl,(sp)		; hl = n1, de = shift
	;jp	sru
	
; ________________________________
; shift right unsigned
;
; sru: hl = hl >> e

sru::	ld	a,e
	and	a,$F
	ret	z

	cp	a,8
	jr	c,sr2

sr1:	ld	l,h
	ld	h,0
	ret	z
	sub	a,8

sr2:	srl	h
	rr	l
	dec	a
	jr	nz,sr2
	ret

; ________________________________
; signed shift right n1>>n2
;
; SRs ( n1 n2 -- n )

SRs::	ld	hl,exhlde_next
	ex	hl,(sp)		; hl=n1, de=shift
	;jp	srs

; ________________________________
; shift right signed
;
; srs: hl = hl >> e

srs::	ld	a,e
	and	a,$F
	ret	z

	cp	a,8
	jr	c,2$

	bit	7,h
	jr	z,sr1

	ld	l,h
	ld	h,0xff

	sub	a,8
	ret	z

2$:	sra	h
	rr	l
	dec	a
	jr	nz,2$
	ret

; ________________________________
; multiply n1*n2
;
; MUL ( n1 n2 -- n )

MUL::	ld	hl,exhlde_next
	ex	hl,(sp)
	;jr	mul_hl_de

; ________________________________
; multiply signed / unsigned
;
; HL = HL * DE		<-- Result is in HL!
;
; pres: BC, DE

mul_hl_de::
	xor 	a
	cp	a,h
	jr	z,lxde	; h=0 => l*de
	cp	a,d
	jr	z,hlxe	; d=0 => hl*e

; HL≥256 && DE≥256:	; => lot of overflow

	ld	a,l
	push	af	; l
	ld	a,h
	call	axel	; hl = h*de, de=pres
	pop	af	; a=l
	jr	axel	; hl = (h*de)<<8 + l*de


hlxe: ex	de,hl	; HL = HL * E
lxde: ld	a,l	; HL = DE * L
axde: ld	l,0	; HL = A * DE
axel::			; HL = L<<8 + A * DE

	scf		; -> set bit 0
	adc	a	; -> cy, z
	jr	c,3$

1$:	add	hl,hl
2$:	add	a	; -> cy, z
	jr	nc,1$
	ret	z	; this was the stopper
3$:	add	hl,hl
	add	hl,de
	jr	2$


; ________________________________
; remainder of unsigned n1/n2
;
; REMu ( n1 n2 -- n )

REMu::	ld	hl,exhlde_next
	ex	hl,(sp)
	jr	divu

; ________________________________
; remainder of signed n1/n2
;
; REMs ( n1 n2 -- n )

REMs::	ld	hl,exhlde_next
	ex	hl,(sp)
	;jp	rems

; ________________________________
; remainder of signed division
; the remainder has the same sign as HL
;
; HL = HL % DE		<-- result is in HL!

rems::	bit	7,h
	jr	z,div0	; HL pos => result pos
	call	div1	; HL neg => divide and negate result

neg_hl:	xor	a
	sub	a,l
	ld	l,a
	sbc	a,a		; nc => 0, cy => 0xff
	sub	a,h
	ld	h,a
	ret

; ________________________________
; divide signed n1/n2
;
; DIVs ( n1 n2 -- n )

DIVs::	ld	hl,next
	ex	hl,(sp)
	;jp	divs

; ________________________________
; signed division
;
; DE = HL / DE

divs::	ld	a,h
	or	d
	jp	p,divu		; both pos.

	ld	a,h
	xor	d
	jp	p,div		; both neg. => result pos.
	call	div		; else divide & neg result

neg_de:	xor	a
	sub	a,e
	ld	e,a
	sbc	a,a		; nc => 0, cy => 0xff
	sub	a,d
	ld	d,a
	ret

div: 	bit	7,h
div1: 	call	nz,neg_hl
div0: 	bit	7,d
	call	nz,neg_de
	jr	divu

; ________________________________
; divide unsigned n1/n2
;
; DIVu ( n1 n2 -- n )

DIVu::	ld	hl,next
	ex	hl,(sp)
	jp	divu

; ________________________________
; unsigned 16-bit division
;
; DE = HL / DE
; HL = HL % DE
;
; If divisor is 0, quotient is set to "infinity", i.e HL = 0xFFFF.

; Two algorithms:
; one assumes divisor < 2^7,
; the second assumes divisor ≥ 2^7;
; Both algorithms "rotate" 24 bits (H,L,A) but roles change.
; choose the applicable one:

divu::	ld	a,e
	and	a,$80
	or	a,d
	jr	nz,morethan7bits

; ________________________________
; unsigned division
; divisor < 2^7
;
; divu7:	DE = HL / DE  &  HL = HL % DE	 	unsigned	if de ≤ 0x80

divu7::
	ld	d,16		; bits in dividend and possible quotient
	add	hl,hl

dvloop7:
	; HL holds both dividend and quotient. While we shift a bit from
	; MSB of dividend, we shift next bit of quotient in from carry.
	; A holds remainder.
	rla

	; If remainder is ≥ divisor, next bit of quotient is 1:
	sub	a,e		; try to subtract
	jr	nc,nodrop7	; success
	add	a,e       	; else undo, sets cy because sub a,e did set it

nodrop7:
	ccf			; cy = next bit of quotient
	adc	hl,hl		; cy = next bit from current remainder of divident
	dec	d		; preserves cy
	jr	nz,dvloop7

	; Carry now contains the same value it contained before
	; entering dvloop7[*]: "0" = valid result.
	ld	e,a		; DE = remainder, HL = quotient
	ex	de,hl		; DE = quotient,  HL = remainder
	ret

; ________________________________
; unsigned division
; divisor ≥ 2^7

morethan7bits:
	push	bc
	ld	b,9		; at most 9 bits in quotient.
	ld	a,l		; precompute the first 7 shifts
	ld	l,h		; by doing 8
	ld	h,0
	rr	l		; and undoing 1

dvloop:
	; Shift next bit of quotient into bit 0 of dividend
	; Shift next MSB of dividend into LSB of remainder
	; A holds both dividend and quotient. While we shift a bit from
	; MSB of dividend, we shift next bit of quotient in from carry
	; HL holds remainder
	adc	hl,hl		; HL < 2^(7+9), no carry, ever.

	; If remainder is >= divisor, next bit of quotient is 1.
	sbc	hl,de		; try to subtract
	jr	nc,nodrop	; success
	add	hl,de		; else undo
				; this also sets the carry, because sbc hl,de did set it.
nodrop:
	ccf			; cy = next bit of quotient
	rla
	djnz    dvloop

	; Take care of the ninth quotient bit! after the loop B=0.
	rl	b		; BA = quotient; also clears cy
	ld	d,b
	ld	e,a		; DE = quotient, HL = remainder
	pop	bc		; restore
	ret			; nc




; -----------------------------------------
; 	Unsigned Compare
; -----------------------------------------

; ______________________________________
; unsigned compare: n1>n2
;
; GTu ( n1 n2 -- n )

GTu::	pop	hl
	ex	de,hl		; hl=n2, de=n1	
	db	opcode(ld a,N)	; jr ltu

; ______________________________________
; unsigned compare: n1<n2
;
; LTu ( n1 n2 -- n )

LTu::	pop	hl		; n1
ltu:	xor	a		; a=0, cy=0
	sbc	hl,de		; cy => n1-n2<0 => n1<n2 => lt=yes

	ld	d,a		; d=0
	adc	a,a		; nc=>0, cy=>1
	ld	e,a		; de=0|1
	next

; ______________________________________
; unsigned compare: n1≤n2
;
; LEu ( n1 n2 -- n )

LEu::	pop	hl
	ex	de,hl		; hl=n2, de=n1	
	db	opcode(ld a,N)	; jr geu

; ______________________________________
; unsigned compare: n1≥n2
;
; GEu ( n1 n2 -- n )

GEu::	pop	hl
geu:	xor	a		; a=0, cy=0
	sbc	hl,de		; n1-n2

ld_de_nc_next::			; in: a=0 !!!
	ccf			; cy => n1-n2<0 => n1<n2 => ge=no => ccf
	
ld_de_cy_next::			; in: a=0 !!!
	ld	d,a		; d=0
	adc	a,a		; nc=>0, cy=>1
	ld	e,a		; de=0|1
	next

; ______________________________________
; compare two values: n1 == n2
; either signed - signed or
; unsigned - unsigned
;
; EQ ( n n -- n )

EQ::	pop	hl
eq:	xor	a
	sbc	hl,de
	jr	z,return_true

return_false::
	ld	de,0
	next

; ______________________________________
; compare two values: n1 != n2
; either signed - signed or
; unsigned - unsigned
;
; NE ( n n -- n )

NE::	pop	hl
ne:	xor	a
	sbc	hl,de
	jp	z,exhlde_next	; 0

return_true::
	ld	de,1
	next


; -----------------------------------------
; 	Signed Compare
; -----------------------------------------

; _______________________________________
; signed compare: n1>n2
;
; GTs ( n1 n2 -- n )

GTs::	pop	hl		; hl=n1, de=n2, op: hl>de
	ex	hl,de		; op: hl<de
	db	opcode(ld a,N)	; jr lts

; _______________________________________
; signed compare: n1<n2
;
; LTs ( n1 n2 -- n )

LTs::	pop	hl		; hl=n1, de=n2, op: hl<de
lts:	ld	a,h
	xor	d
	jp	p,ltu		; both neg / both pos => use LTu

	xor	a
	add	hl,hl		; cy = hl.vz
	jr	ld_de_cy_next	; result true if hl<0

; _______________________________________
; signed compare: n1≤n2
;
; LEs ( n1 n2 -- n )

LEs::	pop	hl
	ex	de,hl		; hl=n2, de=n1
	db	opcode(ld a,N)	; jr ges
	
; _______________________________________
; signed compare: n1≥n2
;
; GEs ( n1 n2 -- n )

GEs::	pop	hl		; hl=n1, de=n2, op: hl≥de
ges:	ld	a,h
	xor	d
	jp	p,geu		; both neg / both pos => use GEu

	xor	a
	add	hl,hl		; cy = hl.vz
	jr	ld_de_nc_next	; result false if hl<0


; -----------------------------------------
; 	Mixed Signedness Compare
; -----------------------------------------

; ______________________________________
; compare two values with mixed signedness
; either signed - unsigned or vice versa
; 
; EQsu ( n n -- f )
; EQus ( n n -- f )

EQsu::
EQus::	pop	hl
	ld	a,d
	or	h		; either of both exceeds common range?
	jp	p,eq		; no
	jr	return_false	; yes => can't be equal

; ______________________________________
; compare two values with mixed signedness
; either signed - unsigned or vice versa
; 
; NEsu ( n n -- f )
; NEus ( n n -- f )

NEsu::
NEus::	pop	hl
	ld	a,d
	or	h		; either of both exceeds common range?
	jp	p,ne		; no
	jr	return_true	; yes => must be not equal

; ______________________________________
; compare two values with mixed signedness u16 > i16
;
; GTus ( u16 i16 -- f )

GTus::	pop	hl
	ex	hl,de
	db	opcode(cp a,N)	; jr ltsu
	
; ______________________________________
; compare two values with mixed signedness i16 < u16
;
; LTsu ( i16 u16 -- f )

LTsu::	pop	hl
ltsu:	ld	a,h		; i16 < 0 ?
	or	d		; u16 ≥ 0x8000 ?
	jp	p,ltu		; no
	jr	return_true	; yes => i16 is less than u16
	


; ______________________________________
; compare two values with mixed signedness i16 > u16
;
; GTsu ( i16 u16 -- f )

GTsu::	pop	hl
	ex	hl,de
	db	opcode(cp a,N)	; jr ltus
	
; ______________________________________
; compare two values with mixed signedness u16 < i16
;
; LTus ( u16 i16 -- f )

LTus::	pop	hl
ltus:	ld	a,h		; u16 ≥ 0x8000 ?
	or	d		; i16 < 0 ?
	jp	p,ltu
	jr	return_false	; yes => u16 can't be less than i16

; ______________________________________
; compare two values with mixed signedness u16 ≤ i16
;
; LEus ( u16 i16 -- f )

LEus::	pop	hl
	ex	hl,de
	db	opcode(cp a,N)	; jr gesu
	
; ______________________________________
; compare two values with mixed signedness i16 ≥ u16
;
; GEsu ( i16 u16 -- f )

GEsu::	pop	hl
gesu:	ld	a,h		; i16 < 0 ?
	or	d		; u16 ≥ 0x8000 ?
	jp	p,geu		; no
	jr	return_false	; yes => i16 can't be ≥ u16

; ______________________________________
; compare two values with mixed signedness i16 ≤ u16
;
; LEsu ( i16 u16 -- f )

LEsu::	pop	hl
	ex	hl,de
	db	opcode(cp a,N)	; jr geus
	
; ______________________________________
; compare two values with mixed signedness u16 ≥ i16
;
; GEus ( u16 i16 -- f )

GEus::	pop	hl	
geus:	ld	a,h		; u16 ≥ 0x8000 ?
	or	d		; i16 < 0 ?
	jp	p,geu		; no
	jr	return_true	; yes => u16 must be ≥ i16

	






; -----------------------------------------
; Assignments
; -----------------------------------------


; _________________________________
; reset bits in variable
;
; ANDGL ( n &w -- )

ANDGL::	ex	hl,de		; hl->var
	pop	de		; de=value

	ld	a,(hl)
	and	e
	ld	(hl++),a
	ld	a,(hl)
	and	d
	ld	(hl),a

	jp	pop_de_next

; _________________________________
; set bits in variable
;
; ORGL ( n &w -- )

ORGL::	ex	hl,de		; hl->var
	pop	de		; de=value

	ld	a,(hl)
	or	e
	ld	(hl++),a
	ld	a,(hl)
	or	d
	ld	(hl),a

	jp	pop_de_next

; _________________________________
; toggle bits in variable
;
; XORGL ( n &w -- )

XORGL::	ex	hl,de		; hl->var
	pop	de		; de=value

	ld	a,(hl)
	xor	e
	ld	(hl++),a
	ld	a,(hl)
	xor	d
	ld	(hl),a

	jp	pop_de_next

; _________________________________
; add value to variable
;
; ADDGL ( n &w -- )

ADDGL::	ex	hl,de		; hl->var
	pop	de		; de=value

	ld	a,(hl)
	add	e
	ld	(hl++),a
	ld	a,(hl)
	adc	d
	ld	(hl),a
	
	jp	pop_de_next

; _________________________________
; subtract value from variable
;
; SUBGL ( n &w -- )

SUBGL::	ex	hl,de		; hl->var
	pop	de		; de=value

	ld	a,(hl)
	sub	e
	ld	(hl++),a
	ld	a,(hl)
	sbc	d
	ld	(hl),a

	jp	pop_de_next

; _________________________________
; divide unsigned variable
;
; DIVGLu ( n &w -- )

DIVGLu::
	ex	hl,de		; hl -> var

	ld	e,(hl++)
	ld	d,(hl)		; de = var, hl->var.hi
	
	ex	hl,(sp)		; de=var, hl=n, stack:->var.hi

	ex	hl,de	
	call	divu		; de = hl / de

	jr	pop_hl_store_de_xhl_and_drop

; _________________________________
; divide signed variable
;
; DIVGLs ( n &w -- )

DIVGLs::
	ex	hl,de		; hl -> var

	ld	e,(hl++)
	ld	d,(hl)		; de = var, hl->var.hi
	
	ex	hl,(sp)		; de=var, hl=n, stack:->var.hi

	ex	hl,de	
	call	divs		; de = hl / de
	
	jr	pop_hl_store_de_xhl_and_drop

; _________________________________
; divide unsigned variable and store remainder
;
; REMGLu ( n &w -- )

REMGLu::
	ex	hl,de		; hl -> var

	ld	e,(hl++)
	ld	d,(hl)		; de = var, hl->var.hi
	
	ex	hl,(sp)		; de=var, hl=n, stack:->var.hi

	ex	hl,de	
	call	divu		; hl = hl % de

	jr	exhlde_pop_hl_store_de_xhl_and_drop

; _________________________________
; divide signed variable and store remainder
;
; REMGLs ( n &w -- )

REMGLs::
	ex	hl,de		; hl -> var

	ld	e,(hl++)
	ld	d,(hl)		; de = var, hl->var.hi
	
	ex	hl,(sp)		; de=var, hl=n, stack:->var.hi

	ex	hl,de	
	call	rems		; hl = hl % de
	
	jr	exhlde_pop_hl_store_de_xhl_and_drop

; _________________________________
; multiply value to variable
;
; MULGL ( n &w -- )

MULGL::
	ex	hl,de		; hl -> var

	ld	e,(hl++)
	ld	d,(hl)		; de = var, hl->var.hi
	
	ex	hl,(sp)		; de=var, hl=n, stack:->var.hi
	
	call	mul_hl_de	; hl = hl * de	

exhlde_pop_hl_store_de_xhl_and_drop:
	ex	hl,de		; de = result

pop_hl_store_de_xhl_and_drop:
	pop	hl		; hl -> var.hi
	
	ld	(hl),d
	ld	(--hl),e
	
	pop	de
	next

; _________________________________
; shift left variable
;
; SLGL ( n4 &w -- )

SLGL::	ex	hl,de		; hl -> var
	
	ld	e,(hl++)
	ld	d,(hl)		; de = var
	
	ex	hl,(sp)		; stack:->var, hl = shift, de = var

;	call	sl		; de = de << l
	ex	hl,de
	call	sl		; hl = hl << e
	
	jr	exhlde_pop_hl_store_de_xhl_and_drop

; _________________________________
; shift right signed variable
;
; SRGLs ( n4 &w -- )

SRGLs::	ex	hl,de		; hl -> var
	
	ld	e,(hl++)
	ld	d,(hl)		; de = var
	
	ex	hl,(sp)		; stack:->var, hl = shift, de = var

	;call	srs		; de = de << l
	ex	hl,de
	call	srs		; hl = hl >> e
	
	jr	exhlde_pop_hl_store_de_xhl_and_drop

; _________________________________
; shift right unsigned variable
;
; SRGLu ( n4 &w -- )

SRGLu::	ex	hl,de		; hl -> var
	
	ld	e,(hl++)
	ld	d,(hl)		; de = var
	
	ex	hl,(sp)		; stack:->var, hl = shift, de = var

	;call	sru		; de = de << l
	ex	hl,de
	call	sru		; hl = hl >> e
	
	jr	exhlde_pop_hl_store_de_xhl_and_drop

; ________________________________
; decrement word pointer variable
;
; MMSETp16 ( &w* -- )		

MMSETp16::
	ld	a,(de)
	sub	2
	ld	(de),a
	jp	nc,pop_de_next
	ex	hl,de
	inc	hl
	dec	(hl)
	jp	pop_de_next

; ________________________________
; increment word pointer variable
;
; PPSETp16 ( &w* -- )		

PPSETp16::
	ld	a,(de)
	add	2
	ld	(de),a
	jp	nc,pop_de_next
	ex	hl,de
	inc	hl
	inc	(hl)
	jp	pop_de_next

; ________________________________
; increment variable
;
; PPSET   ( &w -- )
; PPSETp8 ( &b* -- )		; increment byte pointer

PPSETp8::
PPSET::	ex	de,hl		; hl->var
	
ppset:	inc	(hl)
	jp	nz,pop_de_next
	inc	hl
	inc	(hl)

	jp	pop_de_next

; ________________________________
; decrement variable
;
; MMSET   ( &w -- )		; decrement word variable
; MMSETp8 ( &b* -- )		; decrement byte pointer

MMSETp8::
MMSET::	ex	de,hl		; hl->var
	
	dec	(hl)
	ld	a,(hl)
	inc	a
	jp	nz,pop_de_next
	inc	hl
	dec	(hl)
	
	jp	pop_de_next

; ________________________________
; get value from variable
; with post-increment
;
; PEEKPP   ( &w -- n )
; PEEKPPp8 ( &b* -- b* )	; peek byte pointer with post increment

PEEKPPp8::
PEEKPP::
	ex	de,hl		; hl->var

	ld	e,(hl++)
	ld	d,(hl)
	inc	de
	ld	(hl),d
	ld	(--hl),e

	dec	de
	jp	next

; ________________________________
; get value from variable
; with post-decrement
;
; PEEKMM   ( &w -- n )
; PEEKMMp8 ( &b* -- b* )	; peek byte pointer with post decrement

PEEKMMp8::
PEEKMM::
	ex	de,hl		; hl->var

	ld	e,(hl++)
	ld	d,(hl)
	dec	de
	ld	(hl),d
	ld	(--hl),e

	inc	de
	jp	next

; ________________________________
; get value from variable
; with pre-decrement
;
; MMPEEK   ( &w -- n )
; MMPEEKp8 ( &b* -- b* )	; peek bype pointer with pre decrement

MMPEEKp8::
MMPEEK::
	ex	de,hl		; hl->var

	ld	e,(hl++)
	ld	d,(hl)
	dec	de
	ld	(hl),d
	ld	(--hl),e

	jp	next

; ________________________________
; get value from variable
; with pre-increment
;
; PPPEEK   ( &w -- n )
; PPPEEKp8 ( &b* -- b* )	; peek byte pointer with pre-increment

PPPEEKp8::
PPPEEK::
	ex	de,hl		; hl->var

	ld	e,(hl++)
	ld	d,(hl)
	inc	de
	ld	(hl),d
	ld	(--hl),e

	jp	next

; ________________________________
; get value from variable
;
; PEEK ( &w -- n )

PEEK::	ex	hl,de		; hl -> var	
	jp	ld_de_xhl_next

; ________________________________
; store value in variable
;
; POKE ( n &w -- )

POKE::	ex	hl,de		; hl -> var
	pop	de	
	jp	ld_xhl_de_drop

; ________________________________
; store value in variable
;
; NPOKE ( &w n -- )

NPOKE::	pop	hl		; hl -> var
	jp	ld_xhl_de_drop
	


; ________________________________
; 16 bit Galois LFSR
;
; https://en.wikipedia.org/wiki/Linear_feedback_shift_register
;
; this example produces the bit stream 
; the used polynomial results in a period of 2^16-1
; IMPORTANT: seed must not be 0x0000
; 	     0x0000 is never returned!
;
; uint16 start_state = 0xACE1u;  /* Any nonzero start state will work. */
; uint16 lfsr = start_state;
; unsigned period = 0;
;
; do	
; {
;     unsigned lsb = lfsr & 1;   /* Get LSB (i.e., the output bit). */
;     lfsr >>= 1;                /* Shift register */
;     lfsr ^= (-lsb) & 0xB400u;  /* If the output bit is 1, apply toggle mask.
;                                 * The value has 1 at bits corresponding
;                                 * to taps, 0 elsewhere. */
;     ++period;
; } while (lfsr != start_state);
;
; wir schieben andersrum und das Polynom ist = $002D
; 0xB400 = 10110100 00000000 -> 00000000 00101101 = 0x002D
;
; RAND ( -- n )

RAND::	push	de
	
	ld	hl,(rndseed16)	; get LFSR
	
	ld	de,0x2D10	; d = polynomial.lo, e = counter for 16 bits
		
2$:	add	hl,hl
	jr	nc,1$
	ld	a,l
	xor	d
	ld	l,a
1$:	dec	e
	jp	nz,2$
		
	ld	(rndseed16),hl	; save LFSR
	
	ex	hl,de
	dec	de		; range [0xffff..0x0001] -> [0xfffe..0x0000]
	jp	next


; ________________________________
; swap two word variables
;
; SWAPVARS ( &var &var -- )

SWAPVARS::
	pop	hl

swapvars::	
	ld	a,(de)
	ld	e,(hl)
	ld	(hl++),a
	ld	a,e
	ld	(de++),a

	ld	a,(de)
	ld	e,(hl)
	ld	(hl),a
	ld	a,e
	ld	(de),a
	
	jp	pop_de_next



#endlocal

